/*****
*
*   life.cs
*
*   copyright 2002, Kevin Lindsey
*   licensing information available at: http://www.kevlindev.com/license.txt
*
*****/
using System;
using System.Drawing;
using System.Windows.Forms;

public class Life : Form {
    static int[,] Rules = {
        //0  1  2  3  4  5  6  7  8
        { 0, 0, 0, 1, 0, 0, 0, 0, 0 },
        { 0, 0, 1, 1, 0, 0, 0, 0, 0 }
    };

    Timer timer;
    StatusBar statusBar;
    Bitmap bm;

    int cellSize = 3;
    int cellPad  = 1;

    int generation;
    int buffer;
    int width;
    int height;
    int maxLives;
    SolidBrush[] brush;
    int[,,] Cells;

    MenuItem StartMenuItem;
    MenuItem StopMenuItem;
    MenuItem AdvanceMenuItem;
    
    
    /*****
    *
    *   Main
    *
    *****/
    [System.STAThreadAttribute()]
    public static void Main() {
        Application.Run( new Life(100,100, 100) );
    }


    /*****
    *
    *   constructor
    *
    *****/
    public Life(int width, int height, int maxLives) {
        this.width    = width;
        this.height   = height;
        this.Cells    = new int[2,width,height];

        this.maxLives = maxLives;
        this.brush    = new SolidBrush[maxLives + 1];

        this.timer = new Timer();

        this.bm = new Bitmap( (cellSize+cellPad)*width, (cellSize+cellPad)*height);

        // init colors
        InitializeColors();

        // init component
        InitializeComponent(
            (cellSize + cellPad) * width,
            (cellSize + cellPad) * height
        );

        // init world
        InitializeCells();

        // start timer
        timer.Tick += new EventHandler(TimerHandler);
        timer.Interval = 30;
        Start();
    }


    /*****
    *
    *   InitializeColors
    *
    *   RGB color1 = new RGB(0,0,0);
    *   RGB color2 = new RGB(255,255,255);
    *
    *****/
    private void InitializeColors() {
        RGB color1 = new RGB(192,0,0);
        RGB color2 = new RGB(255,255,0);

        brush[0] = new SolidBrush(Color.Black);
        
        for ( int life = 1; life <= maxLives; life++ ) {
            RGB color3 = color1.lerp(color2, (float) life / maxLives);
            
            brush[life] = new SolidBrush(
                Color.FromArgb(255, color3.r, color3.g, color3.b)
            );
        }
    }


    /*****
    *
    *   InitializeCells
    *
    *****/
    private void InitializeCells() {
        Random gen = new Random();
        Graphics g = Graphics.FromImage(bm);

        this.buffer      = 0;
        this.generation  = 0;

        for ( int x = 0; x < width; x++ ) {
            for ( int y = 0; y < height; y++ ) {
                Cells[buffer,x,y] = ( gen.Next(100) >= 75 ) ? maxLives : 0;
            }
        }

        // Init world graphics
        g.Clear(Color.Black);
        PaintCells();
    }


    /*****
    *
    *   InitializeComponent
    *
    *****/
    private void InitializeComponent(int width, int height) {
        // adjust window appearance
        this.SetStyle(
            ControlStyles.AllPaintingInWmPaint | ControlStyles.Opaque | ControlStyles.UserPaint,
            true
        );
        this.ClientSize      = new System.Drawing.Size(width+1,height+1);
        this.Text            = "Life";
        this.Paint           += new PaintEventHandler(this.PaintHandler);
        this.MaximizeBox     = false;
        this.MinimizeBox     = false;
        this.FormBorderStyle = FormBorderStyle.FixedDialog;
        //this.SizeGripStyle   = SizeGripStyle.Hide;
        //this.MaximumSize = new Size(width, height);
        //this.MinimumSize = new Size(width, height);

        // create menu
        this.Menu = new MainMenu();

        MenuItem ClearMenuItem, RandomMenuItem, GlidersMenuItem, ExitMenuItem;
        MenuItem mi = this.Menu.MenuItems.Add("&Run");

        StartMenuItem   = new MenuItem("&Start", new EventHandler(StartHandler), Shortcut.CtrlS);
        StopMenuItem    = new MenuItem("S&top", new EventHandler(StopHandler), Shortcut.CtrlT);
        AdvanceMenuItem = new MenuItem("&Advance", new EventHandler(AdvanceHandler), Shortcut.CtrlA);
        mi.MenuItems.Add( StartMenuItem );
        mi.MenuItems.Add( StopMenuItem );
        mi.MenuItems.Add( AdvanceMenuItem );

        mi.MenuItems.Add("-");
        ClearMenuItem   = new MenuItem("&Clear", new EventHandler(ClearHandler), Shortcut.CtrlC);
        RandomMenuItem  = new MenuItem("&Random", new EventHandler(RandomHandler), Shortcut.CtrlR);
        GlidersMenuItem = new MenuItem("&Gliders", new EventHandler(GlidersHandler), Shortcut.CtrlG);
        mi.MenuItems.Add( ClearMenuItem );
        mi.MenuItems.Add( RandomMenuItem );
        mi.MenuItems.Add( GlidersMenuItem );
        mi.MenuItems.Add("-");

        ExitMenuItem = new MenuItem("E&xit", new EventHandler(ExitHandler), Shortcut.CtrlX);
        mi.MenuItems.Add( ExitMenuItem );

        // create status bar
        statusBar = new StatusBar();
        statusBar.BackColor = SystemColors.Control;
        statusBar.Location  = new Point(0, 180);
        statusBar.Size      = new Size(450, 20);
        statusBar.TabIndex  = 1;
        statusBar.Text      = "Stopped";
        Controls.Add(statusBar);
    }


    /*****
    *
    *   Event Handlers
    *
    *****/
    /*****
    *
    *   TimerHandler
    *
    *****/
    private void TimerHandler(Object obj, EventArgs args) {
        Advance();
    }

    
    /*****
    *
    *   PaintHandler
    *
    *****/
    private void PaintHandler(object send, PaintEventArgs e) {
        Graphics g = e.Graphics;

        g.DrawImage(bm, 0, 0);
    }


    /*****
    *
    *   PaintCells
    *
    *****/
    private void PaintCells() {
        Graphics g = Graphics.FromImage(bm);

        for ( int x = 0; x < width; x++ ) {
            for ( int y = 0; y < height; y++ ) {
                g.FillRectangle(
                    brush[ Cells[buffer,x,y] ],
                    (cellSize + cellPad) * x + 1, (cellSize + cellPad) * y + 1,
                    cellSize, cellSize
                );
            }
        }
    }


    /*****
    *
    *   StartHandler
    *
    *****/
    private void StartHandler(Object obj, EventArgs e) {
        Start();
    }

    
    /*****
    *
    *   StopHandler
    *
    *****/
    private void StopHandler(Object obj, EventArgs e) {
        Stop();
    }


    /*****
    *
    *   AdvanceHandler
    *
    *****/
    private void AdvanceHandler(Object obj, EventArgs e) {
        Advance();
    }

    
    /*****
    *
    *   RandomHandler
    *
    *****/
    private void RandomHandler(Object obj, EventArgs e) {
        // re-initialize world
        InitializeCells();

        // force redraw
        Invalidate();
    }


    /*****
    *
    *   ClearHandler
    *
    *****/
    private void ClearHandler(Object obj, EventArgs e) {
        // stop timer in case it's running
        Stop();

        // re-initialize world
        this.buffer      = 0;
        this.generation  = 0;

        for ( int x = 0; x < width; x++ ) {
            for ( int y = 0; y < height; y++ ) {
                Cells[buffer,x,y] = 0;
            }
        }

        PaintCells();

        // force redraw
        Invalidate();
    }


    /*****
    *
    *   GlidersHandler
    *
    *****/
    private void GlidersHandler(Object obj, EventArgs e) {
        Random gen = new Random();
        
        for ( int i = 0; i < 10; i++ ) {
            PutGlider( gen.Next(width), gen.Next(height), gen.Next(4) );
        }

        PaintCells();

        // force redraw
        Invalidate();
    }

    
    /*****
    *
    *   ExitHandler
    *
    *****/
    private void ExitHandler(Object obj, EventArgs e) {
        Application.Exit();
    }


    /*****
    *
    *   PutGlider
    *
    *****/
    private void PutGlider(int x, int y, int direction) {
        x = x % width;
        y = y % height;

        int xm1 = (x - 1 + width) % width;
        int xm2 = (x - 2 + width) % width;
        int xp1 = (x + 1) % width;
        int xp2 = (x + 2) % width;

        int ym1 = (y - 1 + height) % height;
        int ym2 = (y - 2 + height) % height;
        int yp1 = (y + 1) % height;
        int yp2 = (y + 2) % height;

        switch (direction) {
            case 0:
                Cells[buffer, x,   y  ] = maxLives;
                Cells[buffer, xp1, y  ] = maxLives;
                Cells[buffer, xp2, y  ] = maxLives;
                Cells[buffer, x,   yp1] = maxLives;
                Cells[buffer, xp1, yp2] = maxLives;
                break;

            case 1:
                Cells[buffer, x,   y  ] = maxLives;
                Cells[buffer, x,   ym1] = maxLives;
                Cells[buffer, x,   ym2] = maxLives;
                Cells[buffer, xp1, y  ] = maxLives;
                Cells[buffer, xp2, ym1] = maxLives;
                break;

            case 2:
                Cells[buffer, x,   y  ] = maxLives;
                Cells[buffer, xm1, y  ] = maxLives;
                Cells[buffer, xm2, y  ] = maxLives;
                Cells[buffer, x,   ym1] = maxLives;
                Cells[buffer, xm1, ym2] = maxLives;
                break;

            case 3:
                Cells[buffer, x,   y  ] = maxLives;
                Cells[buffer, x,   yp1] = maxLives;
                Cells[buffer, x,   yp2] = maxLives;
                Cells[buffer, xm1, y  ] = maxLives;
                Cells[buffer, xm2, yp1] = maxLives;
                break;
        }
    }


    /*****
    *
    *   Start
    *
    *****/
    private void Start() {
        StartMenuItem.Enabled   = false;
        StopMenuItem.Enabled    = true;
        AdvanceMenuItem.Enabled = false;
        statusBar.Text = "Running";
        timer.Start();
    }


    /*****
    *
    *   Stop
    *
    *****/
    private void Stop() {
        StartMenuItem.Enabled   = true;
        StopMenuItem.Enabled    = false;
        AdvanceMenuItem.Enabled = true;
        statusBar.Text = "Stopped";
        timer.Stop();
    }


    /*****
    *
    *   Advance
    *
    *****/
    private void Advance() {
        int oldGen = buffer;

        generation++;
        buffer = 1 - buffer;

        for ( int x = 0; x < width; x++ ) {
            for ( int y = 0; y < height; y++ ) {
                int value = 0;
                int newValue;
                
                int xm1 = (x + width  - 1) % width;
                int xp1 = (x + width  + 1) % width;
                int ym1 = (y + height - 1) % height;
                int yp1 = (y + height + 1) % height;

                value += ( Cells[oldGen, xm1, ym1] == maxLives ) ? 1 : 0;
                value += ( Cells[oldGen, x,   ym1] == maxLives ) ? 1 : 0;
                value += ( Cells[oldGen, xp1, ym1] == maxLives ) ? 1 : 0;
                value += ( Cells[oldGen, xm1, y  ] == maxLives ) ? 1 : 0;
                value += ( Cells[oldGen, xp1, y  ] == maxLives ) ? 1 : 0;
                value += ( Cells[oldGen, xm1, yp1] == maxLives ) ? 1 : 0;
                value += ( Cells[oldGen, x,   yp1] == maxLives ) ? 1 : 0;
                value += ( Cells[oldGen, xp1, yp1] == maxLives ) ? 1 : 0;
                
                if ( Cells[oldGen, x, y] == maxLives ) {
                    newValue = Rules[1, value];
                } else {
                    newValue = Rules[0, value];
                }

                if ( newValue == 1 ) {
                    Cells[buffer, x, y] = maxLives;
                } else {
                    if ( Cells[oldGen, x, y] > 0 ) {
                        Cells[buffer, x, y] = Cells[oldGen, x, y] - 1;
                    } else {
                        Cells[buffer, x, y] = 0;
                    }
                }
            }
        }
        
        PaintCells();
        Invalidate();
    }
}


/*****
*
*   RGB
*
*   helper class
*
*****/
public class RGB {
    public int r, g, b;
    
    public RGB(int r, int g, int b) {
        this.r = r;
        this.g = g;
        this.b = b;
    }

    public RGB lerp(RGB that, float percent) {
        return new RGB(
            (int) (r + (that.r - r) * percent),
            (int) (g + (that.g - g) * percent),
            (int) (b + (that.b - b) * percent)
        );
    }
}
